#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"
#include "xmltok/xmlparse/xmlparse.h"

typedef struct {
    SV* self_sv;
    SV* start_sv;
    SV* end_sv;
    SV* char_sv;
    SV* proc_sv;
    I32 flags;
} CallbackVector;

static void characterData(void *userData, const char *s, int len)
{
  dSP;
  CallbackVector* v = (CallbackVector*) userData;

  PUSHMARK(sp);
  XPUSHs(v->self_sv);
  XPUSHs(sv_2mortal(newSVpv((char*)s,len)));
  PUTBACK;
  perl_call_sv(v->char_sv, v->flags);
}

static void startElement(void *userData, const char *name, const char **atts)
{
  dSP;
  CallbackVector* v = (CallbackVector*) userData;

  PUSHMARK(sp);
  XPUSHs(v->self_sv);
  XPUSHs(sv_2mortal(newSVpv((char*)name,0)));
  while (*atts)
      XPUSHs(sv_2mortal(newSVpv((char*)*atts++,0)));
  PUTBACK;
  perl_call_sv(v->start_sv, v->flags);
}

static void endElement(void *userData, const char *name)
{
  dSP;
  CallbackVector* v = (CallbackVector*) userData;

  PUSHMARK(sp);
  XPUSHs(v->self_sv);
  XPUSHs(sv_2mortal(newSVpv((char*)name,0)));
  PUTBACK;
  perl_call_sv(v->end_sv, v->flags);
}

static void processingInstruction(void *userData, const char *target, const char *data)
{
  dSP;
  CallbackVector* v = (CallbackVector*) userData;

  PUSHMARK(sp);
  XPUSHs(v->self_sv);
  XPUSHs(sv_2mortal(newSVpv((char*)target,0)));
  XPUSHs(sv_2mortal(newSVpv((char*)data,0)));
  PUTBACK;
  perl_call_sv(v->proc_sv, v->flags);
}

MODULE = XML::Parser	PACKAGE = XML::Parser	PREFIX = XML_

XML_Parser
XML_ParserCreate(encoding, self_sv, start_sv, end_sv, char_sv, proc_sv, flags, raw, userdata)
	char *			encoding
	SV *			self_sv
	SV *			start_sv
	SV *			end_sv
	SV *			char_sv
	SV *			proc_sv
	I32			flags
	int			raw
	void *			userdata
    CODE:
	{
	    CallbackVector *v = (CallbackVector *)malloc(sizeof(CallbackVector));
	    v->self_sv = SvREFCNT_inc(self_sv);	/* XXX makes loop */
	    v->start_sv = start_sv;
	    v->end_sv = end_sv;
	    v->char_sv = char_sv;
	    v->proc_sv = proc_sv;
	    v->flags = flags;
	    RETVAL = XML_ParserCreate(encoding, raw);
	    XML_SetUserData(RETVAL, (void*)v);
	    XML_SetElementHandler(RETVAL, startElement, endElement);
	    XML_SetCharacterDataHandler(RETVAL, characterData);
	    XML_SetProcessingInstructionHandler(RETVAL, processingInstruction);
	    userdata = (void*)v;
	}
    OUTPUT:
	RETVAL
	userdata

void
XML_ParserFree(parser, userdata)
	XML_Parser			parser
	void *				userdata
    CODE:
	XML_ParserFree(parser);
	free(userdata);

#void
#XML_SetElementHandler
#	XML_Parser			parser
#	XML_StartElementHandler		start
#	XML_EndElementHandler		end
#
#void
#XML_SetCharacterDataHandler
#	XML_Parser			parser
#	XML_CharacterDataHandler	handler
#
#void
#XML_SetProcessingInstructionHandler
#	XML_Parser			parser
#	XML_ProcessingInstructionHandler handler
#
#void
#XML_SetUserData
#	XML_Parser			parser
#	void				*userData

int
XML_Parse(parser, s, isFinal)
	XML_Parser			parser
	char *				s
	int				len = na;
	int				isFinal
    CODE:
	RETVAL = XML_Parse(parser,s,len,isFinal);
    OUTPUT:
	RETVAL

#void
#XML_GetBuffer
#	XML_Parser			parser
#	int				len
#
#int
#XML_ParseBuffer
#	XML_Parser			parser
#	int				len
#	int				isFinal

int
XML_GetErrorCode(parser)
	XML_Parser			parser

int
XML_GetErrorLineNumber(parser)
	XML_Parser			parser

int
XML_GetErrorColumnNumber(parser)
	XML_Parser			parser

long
XML_GetErrorByteIndex(parser)
	XML_Parser			parser

char *
XML_ErrorString(code)
	int				code
    CODE:
	RETVAL = (char*) XML_ErrorString(code);
    OUTPUT:
	RETVAL
